; -----------------------------------------------
;  draw-objects.scm - Ben VandenBos & Tim Reeves
;
;  *Functions for drawing and moving objects
; -----------------------------------------------

; Draw Tank Functions
; -------------------

; Constants
(define *barrel-length* 18)        ; Tank Barrelt length (pixels)
(define *tank-height-width* 45)    ; Tank height/width (outer rectangle) (pixels)
(define *turret-height-width* 42)  ; Tank turret heigth/width (inner circle) (pixels)
(define *bullet-width* 5)          ; Bullet width (pixels)
(define *bitmap-color* (instantiate color% ("White")))
(define *x-offset* 0)
(define *y-offset* 0)


; Move Functions
; --------------

; function that loops in a thread, drawing a specific index in the game-state
; This function draws any object based on it's index.  It looks up the proper
; drawing functions for each type in the *move-constants* hash-table the returns
; a struct of constants, including drawing functions...
(define make-object-move
  (lambda (index)
    (semaphore-wait *game-state-semaphore*)
    (let ((obj (vector-ref *game-state* index)))
      (semaphore-post *game-state-semaphore*)
      (if (object? obj)
          (begin
            (letrec ((obj-consts (hash-table-get *move-constants* (object-type obj)))
                     (pos (exact-position obj))
                     (x (car pos))
                     (y (cadr pos))
                     (theta (caddr pos)))
              (if *wire-frame?*
                  ((object-constants-draw-wire-frame obj-consts) index x y theta (object-color obj))
                  ((object-constants-draw-sprite obj-consts) index x y theta (object-color obj)))
              (make-object-move index)
              )
            )
          )   
      )
    )
  )


;(find-bitmap theta) -> vector-index
;returns the vector-position of the bitmap cooresponding to the angle of theta
;such that index => floor (theta/9), where theta is in degrees

(define find-bitmap
  (lambda (theta)
    (let ((deg (modulo (inexact->exact (floor (* theta (/ 180 pi)))) 360)))
      (cond
        ((< deg 0) (set! deg (+ 360 deg))))
      (if (>= deg 360) 0 
          (floor (/ deg 9))))))

; Used to round-off-theta so everything is displayed
; to the nearest 2pi/40 (360/40)
(define round-off-theta
  (lambda (theta)
    (* (* 9 (find-bitmap theta)) (/ pi 180))))

; Draw Tank Functions
; -------------------

;(draw-tank x y theta color) -> void
;Draws the tank according to a theta representation of the tank position
;e.g. if theta => 220 then draw-tank will draw tank216 which is a 216 degree position of the tank 
;that represents a 220 degree position of the tank....

(define draw-sprite-tank
  (lambda (index x y theta color)
    (let ((bitmap-index (find-bitmap theta))
          (height (+ (send drawspace get-height) *y-offset*))
          (width  (+ (send drawspace get-width) *x-offset*))
          (x-offset *x-offset*)
          (y-offset *y-offset*))
      ; First draw what we need to draw...
      (if (eq? index *you-are-index*)
          (begin
            (if (and (> x *pan-thresh*)(< x (+ *x-offset* *pan-thresh*)))
                (set! *x-offset* (- *x-offset* (- (+ *pan-thresh* *x-offset*) x))))            
            (if (and (< x (- 1600 *pan-thresh*))(> x (- width *pan-thresh*)))
                (set! *x-offset* (+ *x-offset* (- x (- width *pan-thresh*)))))
            (if (and (> y *pan-thresh*)(< y (+ *y-offset* *pan-thresh*)))
                (set! *y-offset* (- *y-offset* (- (+ *pan-thresh* *y-offset*) y))))            
            (if (and (< y (- 1200 *pan-thresh*))(> y (- height *pan-thresh*)))
                (set! *y-offset* (+ *y-offset* (- y (- height *pan-thresh*)))))
            (send mem-dc draw-bitmap (vector-ref tank-sprites-vec bitmap-index) 
                  (- x (/ *turret-height-width* 2)) (- y (/ *turret-height-width* 2)) 'solid *bitmap-color* (vector-ref tank-mask-vec bitmap-index)))
          
          (begin
            (if (and (< x x-offset) (< y y-offset))
                (send mem-dc draw-bitmap arrow-upleft (+ 4 x-offset) (+ 4 y-offset) 'solid *bitmap-color* arrow-upleft-mask))
            (if (and (< x x-offset) (< y height))
                (send mem-dc draw-bitmap arrow-left x-offset y 'solid *bitmap-color* arrow-left-mask))
            (if (and (< x x-offset) (> y height))
                (send mem-dc draw-bitmap arrow-downleft (+ 4 x-offset) (- height 40) 'solid *bitmap-color* arrow-downleft-mask))
            (if (and (> x width) (< y y-offset))
                (send mem-dc draw-bitmap arrow-upright (- width 40) (+ 4 y-offset) 'solid *bitmap-color* arrow-upright-mask))
            (if (and (< x width) (< y y-offset))
                (send mem-dc draw-bitmap arrow-up x y-offset 'solid *bitmap-color* arrow-up-mask))
            (if (and (> x width) (< y height))
                (send mem-dc draw-bitmap arrow-right (- width 40) y 'solid *bitmap-color* arrow-right-mask))
            (if (and (< x width) (> y height))
                (send mem-dc draw-bitmap arrow-down x (- height 40) 'solid *bitmap-color* arrow-down-mask))
            (if (and (> x width) (> y height))
                (send mem-dc draw-bitmap arrow-downright (- width 40) (- height 40) 'solid *bitmap-color* arrow-downright-mask)) 
            (send mem-dc draw-bitmap (vector-ref e-tank-sprites-vec bitmap-index) 
                  (- x (/ *turret-height-width* 2)) (- y (/ *turret-height-width* 2)) 'solid *bitmap-color* (vector-ref e-tank-mask-vec bitmap-index))))
      
      (sleep/yield *refresh-delay*)
      
      ; Now draw over whatever we just drew...
      (if (eq? index *you-are-index*)
          (send mem-dc draw-bitmap-section game-board
                (- x (/ *turret-height-width* 2)) (- y (/ *turret-height-width* 2))
                (- x (/ *turret-height-width* 2)) (- y (/ *turret-height-width* 2)) 58 58)
          
          (begin
            (if (and (< x x-offset)(< y y-offset))
                (send mem-dc draw-bitmap-section game-board  (+ 4 x-offset) (+ 4 y-offset) (+ 4 x-offset) (+ 4 y-offset) 36 36))
            (if (and (< x x-offset)(< y height))
                (send mem-dc draw-bitmap-section game-board  x-offset y x-offset y 41 34))
            (if (and (< x x-offset)(> y height))
                (send mem-dc draw-bitmap-section game-board  (+ 4 x-offset) (- height 40) (+ 4 x-offset) (- height 40) 36 36))
            (if (and (> x width)(< y y-offset))
                (send mem-dc draw-bitmap-section game-board  (- width 40) (+ 4 y-offset) (- width 40) (+ 4 y-offset) 36 36))
            (if (and (< x width)(< y y-offset))
                (send mem-dc draw-bitmap-section game-board  x y-offset x y-offset 34 41))
            (if (and (> x width)(< y height))
                (send mem-dc draw-bitmap-section game-board (- width 40) y (- width 40) y 41 34))
            (if (and (< x width)(> y height))
                (send mem-dc draw-bitmap-section game-board x (- height 40) x (- height 40) 34 41))
            (if (and (> x width)(> y height))
                (send mem-dc draw-bitmap-section game-board (- width 40) (- height 40) (- width 40) (- height 40) 36 36))
            (send mem-dc draw-bitmap-section game-board
                  (- x (/ *turret-height-width* 2)) (- y (/ *turret-height-width* 2))
                  (- x (/ *turret-height-width* 2)) (- y (/ *turret-height-width* 2)) 58 58))))))

(define draw-sprite-bullet
  (lambda (index x y theta color)
    (send mem-dc draw-bitmap bullet x y 'solid *bitmap-color* bullet-mask)
    (sleep/yield *refresh-delay*)
    (send mem-dc draw-bitmap-section game-board x y x y 7 7)))

(define draw-wire-tank
  (lambda (index x y theta color)
    (let ((height (+ (send drawspace get-height) *y-offset*))
          (width  (+ (send drawspace get-width) *x-offset*))
          (x-offset *x-offset*)
          (y-offset *y-offset*))
      ; First draw what we need to draw...
      (if (eq? index *you-are-index*)
          (begin
            (if (and (> x *pan-thresh*)(< x (+ *x-offset* *pan-thresh*)))
                (set! *x-offset* (- *x-offset* (- (+ *pan-thresh* *x-offset*) x))))            
            (if (and (< x (- 1600 *pan-thresh*))(> x (- width *pan-thresh*)))
                (set! *x-offset* (+ *x-offset* (- x (- width *pan-thresh*)))))
            (if (and (> y *pan-thresh*)(< y (+ *y-offset* *pan-thresh*)))
                (set! *y-offset* (- *y-offset* (- (+ *pan-thresh* *y-offset*) y))))            
            (if (and (< y (- 1200 *pan-thresh*))(> y (- height *pan-thresh*)))
                (set! *y-offset* (+ *y-offset* (- y (- height *pan-thresh*)))))
            
            (semaphore-wait *drawing-semaphore*)    
            (send mem-wire-dc set-brush *no-brush*)
            (send mem-wire-dc set-pen (instantiate pen% ((symbol->string color) 1 'solid)))
            (send mem-wire-dc draw-ellipse (- x (/ *turret-height-width* 2)) (- y (/ *turret-height-width* 2)) *turret-height-width* *turret-height-width*)
            (send mem-wire-dc draw-line x y (+ x (* *barrel-length* (cos (round-off-theta theta)))) (+ y (* *barrel-length* (sin (round-off-theta theta)))))
            (send mem-wire-dc draw-rectangle (- x (/ *tank-height-width* 2)) (- y (/ *tank-height-width* 2)) *tank-height-width* *tank-height-width*)
            (semaphore-post *drawing-semaphore*))
          
          (begin
            (if (and (< x x-offset) (< y y-offset))
                (begin
                  (semaphore-wait *drawing-semaphore*)
                  (send mem-wire-dc set-pen (instantiate pen% ((symbol->string color) 4 'solid)))
                  (send mem-wire-dc draw-line (+ x-offset 4) (+ y-offset 4) (+ x-offset 4) (+ y-offset 14))
                  (send mem-wire-dc draw-line (+ x-offset 4) (+ y-offset 4) (+ x-offset 14) (+ y-offset 4))
                  (send mem-wire-dc draw-line (+ x-offset 4) (+ y-offset 4) (+ x-offset 24) (+ y-offset 24))
                  (semaphore-post *drawing-semaphore*)))
            (if (and (< x x-offset) (< y height))
                (begin
                  (semaphore-wait *drawing-semaphore*)
                  (send mem-wire-dc set-pen (instantiate pen% ((symbol->string color) 4 'solid)))
                  (send mem-wire-dc draw-line (+ x-offset 4) y (+ x-offset 10) (- y 10))
                  (send mem-wire-dc draw-line (+ x-offset 4) y (+ x-offset 10) (+ y 10))
                  (send mem-wire-dc draw-line (+ x-offset 4) y (+ x-offset 28) y)
                  (semaphore-post *drawing-semaphore*)))
            (if (and (< x x-offset) (> y height))
                (begin
                  (semaphore-wait *drawing-semaphore*)
                  (send mem-wire-dc set-pen (instantiate pen% ((symbol->string color) 4 'solid)))
                  (send mem-wire-dc draw-line (+ x-offset 4) (- height 4) (+ x-offset 4) (- height 14))
                  (send mem-wire-dc draw-line (+ x-offset 4) (- height 4) (+ x-offset 14) (- height 4))
                  (send mem-wire-dc draw-line (+ x-offset 4) (- height 4) (+ x-offset 24) (- height 24))
                  (semaphore-post *drawing-semaphore*)))
            (if (and (> x width) (< y y-offset))
                (begin
                  (semaphore-wait *drawing-semaphore*)
                  (send mem-wire-dc set-pen (instantiate pen% ((symbol->string color) 4 'solid)))
                  (send mem-wire-dc draw-line (- width 4) (+ y-offset 4) (- width 14) (+ y-offset 4))
                  (send mem-wire-dc draw-line (- width 4) (+ y-offset 4) (- width 4) (+ y-offset 14))
                  (send mem-wire-dc draw-line (- width 4) (+ y-offset 4) (- width 24) (+ y-offset 24))
                  (semaphore-post *drawing-semaphore*)))
            (if (and (< x width) (< y y-offset))
                (begin
                  (semaphore-wait *drawing-semaphore*)
                  (send mem-wire-dc set-pen (instantiate pen% ((symbol->string color) 4 'solid)))
                  (send mem-wire-dc draw-line x (+ y-offset 4) (- x 10) (+ y-offset 10))
                  (send mem-wire-dc draw-line x (+ y-offset 4) (+ x 10) (+ y-offset 10))
                  (send mem-wire-dc draw-line x (+ y-offset 4) x (+ y-offset 28))
                  (semaphore-post *drawing-semaphore*)))
            (if (and (> x width) (< y height))
                (begin
                  (semaphore-wait *drawing-semaphore*)
                  (send mem-wire-dc set-pen (instantiate pen% ((symbol->string color) 4 'solid)))
                  (send mem-wire-dc draw-line (- width 4) y (- width 10) (- y 10))
                  (send mem-wire-dc draw-line (- width 4) y (- width 10) (+ y 10))
                  (send mem-wire-dc draw-line (- width 4) y (- width 28) y)
                  (semaphore-post *drawing-semaphore*)))
            (if (and (< x width) (> y height))
                (begin
                  (semaphore-wait *drawing-semaphore*)
                  (send mem-wire-dc set-pen (instantiate pen% ((symbol->string color) 4 'solid)))
                  (send mem-wire-dc draw-line x (- height 4) (- x 10) (- height 10))
                  (send mem-wire-dc draw-line x (- height 4) (+ x 10) (- height 10))
                  (send mem-wire-dc draw-line x (- height 4) x (- height 28))
                  (semaphore-post *drawing-semaphore*)))
            (if (and (> x width) (> y height))
                (begin
                  (semaphore-wait *drawing-semaphore*)
                  (send mem-wire-dc set-pen (instantiate pen% ((symbol->string color) 4 'solid)))
                  (send mem-wire-dc draw-line (- width 4) (- height 4) (- width 14) (- height 4))
                  (send mem-wire-dc draw-line (- width 4) (- height 4) (- width 4) (- height 14))
                  (send mem-wire-dc draw-line (- width 4) (- height 4) (- width 24) (- height 24))
                  (semaphore-post *drawing-semaphore*)))
            (semaphore-wait *drawing-semaphore*)    
            (send mem-wire-dc set-brush *no-brush*)
            (send mem-wire-dc set-pen (instantiate pen% ((symbol->string color) 1 'solid)))
            (send mem-wire-dc draw-ellipse (- x (/ *turret-height-width* 2)) (- y (/ *turret-height-width* 2)) *turret-height-width* *turret-height-width*)
            (send mem-wire-dc draw-line x y (+ x (* *barrel-length* (cos (round-off-theta theta)))) (+ y (* *barrel-length* (sin (round-off-theta theta)))))
            (send mem-wire-dc draw-rectangle (- x (/ *tank-height-width* 2)) (- y (/ *tank-height-width* 2)) *tank-height-width* *tank-height-width*)
            (semaphore-post *drawing-semaphore*))
          )
            
    (sleep/yield *refresh-delay*)
      
      (if (eq? index *you-are-index*)
          (begin
            (semaphore-wait *drawing-semaphore*)    
            (send mem-wire-dc set-brush *white-brush*)
            (send mem-wire-dc set-pen *no-pen*)
            (send mem-wire-dc draw-rectangle (- x (/ *tank-height-width* 2)) (- y (/ *tank-height-width* 2)) *tank-height-width* *tank-height-width*)
            (semaphore-post *drawing-semaphore*))
          (begin
            (if (and (< x x-offset) (< y y-offset))
                (begin
                  (semaphore-wait *drawing-semaphore*)
                  (send mem-wire-dc set-pen (instantiate pen% ((instantiate color% ("White")) 5 'solid)))
                  (send mem-wire-dc draw-line (+ x-offset 4) (+ y-offset 4) (+ x-offset 4) (+ y-offset 14))
                  (send mem-wire-dc draw-line (+ x-offset 4) (+ y-offset 4) (+ x-offset 14) (+ y-offset 4))
                  (send mem-wire-dc draw-line (+ x-offset 4) (+ y-offset 4) (+ x-offset 24) (+ y-offset 24))
                  (semaphore-post *drawing-semaphore*)))
            (if (and (< x x-offset) (< y height))
                (begin
                  (semaphore-wait *drawing-semaphore*)
                  (send mem-wire-dc set-pen (instantiate pen% ((instantiate color% ("White")) 5 'solid)))
                  (send mem-wire-dc draw-line (+ x-offset 4) y (+ x-offset 10) (- y 10))
                  (send mem-wire-dc draw-line (+ x-offset 4) y (+ x-offset 10) (+ y 10))
                  (send mem-wire-dc draw-line (+ x-offset 4) y (+ x-offset 28) y)
                  (semaphore-post *drawing-semaphore*)))
            (if (and (< x x-offset) (> y height))
                (begin
                  (semaphore-wait *drawing-semaphore*)
                  (send mem-wire-dc set-pen (instantiate pen% ((instantiate color% ("White")) 5 'solid)))
                  (send mem-wire-dc draw-line (+ x-offset 4) (- height 4) (+ x-offset 4) (- height 14))
                  (send mem-wire-dc draw-line (+ x-offset 4) (- height 4) (+ x-offset 14) (- height 4))
                  (send mem-wire-dc draw-line (+ x-offset 4) (- height 4) (+ x-offset 24) (- height 24))
                  (semaphore-post *drawing-semaphore*)))
            (if (and (> x width) (< y y-offset))
                (begin
                  (semaphore-wait *drawing-semaphore*)
                  (send mem-wire-dc set-pen (instantiate pen% ((instantiate color% ("White")) 5 'solid)))
                  (send mem-wire-dc draw-line (- width 4) (+ y-offset 4) (- width 14) (+ y-offset 4))
                  (send mem-wire-dc draw-line (- width 4) (+ y-offset 4) (- width 4) (+ y-offset 14))
                  (send mem-wire-dc draw-line (- width 4) (+ y-offset 4) (- width 24) (+ y-offset 24))
                  (semaphore-post *drawing-semaphore*)))
            (if (and (< x width) (< y y-offset))
                (begin
                  (semaphore-wait *drawing-semaphore*)
                  (send mem-wire-dc set-pen (instantiate pen% ((instantiate color% ("White")) 5 'solid)))
                  (send mem-wire-dc draw-line x (+ y-offset 4) (- x 10) (+ y-offset 10))
                  (send mem-wire-dc draw-line x (+ y-offset 4) (+ x 10) (+ y-offset 10))
                  (send mem-wire-dc draw-line x (+ y-offset 4) x (+ y-offset 28))
                  (semaphore-post *drawing-semaphore*)))
            (if (and (> x width) (< y height))
                (begin
                  (semaphore-wait *drawing-semaphore*)
                  (send mem-wire-dc set-pen (instantiate pen% ((instantiate color% ("White")) 5 'solid)))
                  (send mem-wire-dc draw-line (- width 4) y (- width 10) (- y 10))
                  (send mem-wire-dc draw-line (- width 4) y (- width 10) (+ y 10))
                  (send mem-wire-dc draw-line (- width 4) y (- width 28) y)
                  (semaphore-post *drawing-semaphore*)))
            (if (and (< x width) (> y height))
                (begin
                  (semaphore-wait *drawing-semaphore*)
                  (send mem-wire-dc set-pen (instantiate pen% ((instantiate color% ("White")) 5 'solid)))
                  (send mem-wire-dc draw-line x (- height 4) (- x 10) (- height 10))
                  (send mem-wire-dc draw-line x (- height 4) (+ x 10) (- height 10))
                  (send mem-wire-dc draw-line x (- height 4) x (- height 28))
                  (semaphore-post *drawing-semaphore*)))
            (if (and (> x width) (> y height))
                (begin
                  (semaphore-wait *drawing-semaphore*)
                  (send mem-wire-dc set-pen (instantiate pen% ((instantiate color% ("White")) 5 'solid)))
                  (send mem-wire-dc draw-line (- width 4) (- height 4) (- width 14) (- height 4))
                  (send mem-wire-dc draw-line (- width 4) (- height 4) (- width 4) (- height 14))
                  (send mem-wire-dc draw-line (- width 4) (- height 4) (- width 24) (- height 24))
                  (semaphore-post *drawing-semaphore*)))
            
            (semaphore-wait *drawing-semaphore*)    
            (send mem-wire-dc set-brush *white-brush*)
            (send mem-wire-dc set-pen *no-pen*)
            (send mem-wire-dc draw-rectangle (- x (/ *tank-height-width* 2)) (- y (/ *tank-height-width* 2)) *tank-height-width* *tank-height-width*)
            (semaphore-post *drawing-semaphore*)))
    )))

(define draw-wire-bullet
  (lambda (index x y theta color)
    (semaphore-wait *drawing-semaphore*)      
    (send mem-wire-dc set-brush *no-brush*)
    (send mem-wire-dc set-pen (instantiate pen% ((symbol->string color) 1 'solid)))
    (send mem-wire-dc draw-ellipse (- x (/ *bullet-width* 2)) (- y (/ *bullet-width* 2)) *bullet-width* *bullet-width*)
    (semaphore-post *drawing-semaphore*)      

    (sleep/yield *refresh-delay*)

    (semaphore-wait *drawing-semaphore*)  
    (send mem-wire-dc set-brush *white-brush*)
    (send mem-wire-dc set-pen *no-pen*)
    (send mem-wire-dc draw-rectangle (- x (/ *bullet-width* 2)) (- y (/ *bullet-width* 2)) *bullet-width* *bullet-width*)
    (semaphore-post *drawing-semaphore*)    
    ))

(define draw-wire-rubble
  (lambda (index x y theta color)
    (semaphore-wait *drawing-semaphore*)
    (send mem-wire-dc set-brush *no-brush*)
    (send mem-wire-dc set-pen (instantiate pen% ('black 1 'solid)))
    (send mem-wire-dc draw-line (- x (/ *tank-height-width* 2)) (- y (/ *tank-height-width* 2))
          (+ x (/ *tank-height-width* 2)) (+ y (/ *tank-height-width* 2)))
    (send mem-wire-dc draw-line (- x (/ *tank-height-width* 2)) (+ y (/ *tank-height-width* 2))
          (+ x (/ *tank-height-width* 2)) (- y (/ *tank-height-width* 2)))
    (semaphore-post *drawing-semaphore*)      
    
    (sleep/yield *refresh-delay*)
    
    (semaphore-wait *drawing-semaphore*)  
    (send mem-wire-dc set-brush *white-brush*)
    (send mem-wire-dc set-pen *no-pen*)
    (send mem-wire-dc draw-rectangle (- x (/ *tank-height-width* 2)) (- y (/ *tank-height-width* 2)) *tank-height-width* *tank-height-width*)
    (semaphore-post *drawing-semaphore*)  
    ))

(define draw-animation
 (lambda (x y index)
   (if (not (eq? index 17))
       (begin
         (send mem-dc draw-bitmap (vector-ref explode-sprites-vec index)
               (- x 35) (- y 50) 'solid (instantiate color% ("White")) (vector-ref explode-mask-vec index))
         (sleep/yield .05)
         (send mem-dc draw-bitmap-section game-board (- x 35) (- y 50) (- x 35) (- y 50) 72 100)
         (draw-animation x y (+ index 1))))))
    